//*-- Author :  R. Michaels, Apr 2014
//
//////////////////////////////////////////////////////////////////////////
//
// GmpNormAna
//
// Code to summarize the information needed 
// to normalize an experiment and obtain a 
// cross section, including
//
// 1. Prescale factors
// 2. Calibration of charge monitor
// 3. Integrated Charge of a run
// 4. Livetime by trigger type, by helicity, and 
//    an approximate average livetime.
// 5. Number of triggers from bit pattern.
//
// This class requires scalers to be enabled
// in the analyzer.  For scalers, see 
// http://hallaweb.jlab.org/equipment/daq/THaScaler.html
//
// At present the run summary is just a printout,
// eventually one could write to a database.
// This needs some work (hopefully a student will do it).
// In addition, some global variables are
// registered for possible Tree analysis, e.g.
// cutting on helicity-correlated deadtime.
//
// How to use this class:
// 
// 1. Do something like this in the analysis script:
//      GmpNormAna* norm = 
//        new GmpNormAna("N","Normalization Analysis");
//      gHaPhysics->Add(norm);
// 
// 2. Add variables N.* to the output tree.
//
// 3. GmpNormAna will produce an end-of-run printout.
//
//
//////////////////////////////////////////////////////////////////////////

#include "GmpNormAna.h"
#include "THaVarList.h"
#include "THaVar.h"
#include "THaGlobals.h"
#include "THaAnalyzer.h"
#include "THaScalerGroup.h"
#include "THaString.h"
#include "TList.h"
#include "THaEvData.h"
#include "TDatime.h"
#include "THaScaler.h"
#include "TH1.h"
#include "TH2.h"
#include "VarDef.h"
#include <fstream>
#include <iostream>

using namespace std;
using THaString::CmpNoCase;

//_____________________________________________________________________________
GmpNormAna::GmpNormAna( const char* name, const char* descript ) :
  THaPhysicsModule( name, descript )
{
  fSetPrescale = kFALSE;
  fHelEnable = kFALSE;
  myscaler = 0;
  fSpectr = descript;
  normdata = new BNormData();
  // There are NTRIG triggers
  nhit = new Int_t[NTRIG];
  // There are MAXHIT  hits possible on a TDC channel
  tdcdata = new Int_t[MAXHIT*NTRIG];
  t1live = 0;
  hpos_alive = 0;
  hneg_alive = 0;
  herecnt=0;
  ev140_scaler = new Double_t[fgNumRoc * fgNumChan];   
  ev140_plus = new Double_t[fgNumRoc * fgNumChan];   
  ev140_minus = new Double_t[fgNumRoc * fgNumChan];   
  roc_scaler = new Double_t[fgNumRoc * fgNumChan];   
  roc_plus = new Double_t[fgNumRoc * fgNumChan];   
  roc_minus = new Double_t[fgNumRoc * fgNumChan];   
  if (fDEBUG) {
       outputFile = new ofstream("normana.txt");
  } 
 for (int i = 0; i < fgNumRoc*fgNumChan; i++) {
      ev140_scaler[i] = 0; ev140_plus[i] = 0; ev140_minus[i] = 0;
      roc_scaler[i] = 0; roc_plus[i] = 0; roc_minus[i] = 0;
  }
  BookHist();
}

//_____________________________________________________________________________
GmpNormAna::~GmpNormAna()
{

  WriteHist();

  if (fDEBUG) outputFile->close();
  delete normdata;
  delete [] nhit;
  delete [] tdcdata;
  delete [] ev140_scaler;
  delete [] ev140_plus;
  delete [] ev140_minus;
  delete [] roc_scaler;
  delete [] roc_plus;
  delete [] roc_minus;
  for (vector<BRocScaler*>::iterator ir = fRocScaler.begin();
       ir != fRocScaler.end(); ir++) delete *ir;
  SetupRawData( NULL, kDelete ); 


}

//_____________________________________________________________________________
Int_t GmpNormAna::SetupRawData( const TDatime* /* run_time */, EMode mode )
{
  Int_t retval = 0;

  RVarDef vars[] = {
    { "evtypebits", "event type bit pattern",      "evtypebits" },  
    // This first part is based on event type 140 and is "older" code.
    { "bcm",   "bcm ",      "bcm" },
    { "t1live",   "t1 deadtime", "t1live" },
    { "TSdtime",   "Trigger Supervisor scalers deadtime", "TSdtime" },
    { "dlive",   "differential livetime from scaler ",  "dlive" },  
    { "hpos_alive", "helicity positive livetime", "hpos_alive" },  
    { "hneg_alive", "helicity negative livetime", "hneg_alive" },  
    { 0 }
  };

  cout << "setting up RVarDef in Normana "<<endl;

  if( mode != kDefine || !fIsSetup )
    retval = DefineVarsFromList( vars, mode );

  fIsSetup = ( mode == kDefine );
  
  return retval;
}

//_____________________________________________________________________________
Int_t GmpNormAna::End( THaRunBase* ) 
{

  PrintSummary();
  WriteHist();
  return 0;
}


//_____________________________________________________________________________
Int_t GmpNormAna::PrintSummary() const
{
// Print the summary.  Eventually want this to go to a
// database or a file.

  cout << endl << "Summary of Normalization Analysis "<<endl;

  if (myscaler == 0) {
    cout << "Sorry, no scalers.  Therefore GmpNormAna ";
    cout << "failed to work."<<endl;
    return 0;
  }

  cout << "Trigger      Prescale      Num evt     Num in      Livetime"<<endl;
  cout << "---           Factor        CODA file   Scalers"<<endl;

  Int_t nhel=1;
  if (fHelEnable) nhel=3; 
  for (int ihel = 0; ihel < nhel; ihel++) {
    Int_t jhel = normdata->InvHel(ihel);  // jhel = 0, -1, +1
    if (fHelEnable) {
      if (jhel==0) {
        cout << "\nIrrespective of helicity "<<endl;
      } else {
        cout << "\nFor Helicity =  "<<jhel<<endl<<endl;
      }
    }
    for (int itrig = 0; itrig < 8; itrig++) {
// Again be careful about indices.  E.g. GetTrig starts at itrig=1.
      cout << "  "<<itrig+1<<"      "<<normdata->GetPrescale(itrig)<<"      "<<normdata->GetTrigCount(itrig,jhel)<<"       "<<myscaler->GetTrig(jhel,itrig+1)<<"       "<<normdata->GetLive(itrig,jhel)<<endl;

    }
    cout << "\nNumber of events = "<<normdata->GetEvCount(jhel)<<endl;
    cout << "Run-averaged livetime from scalers = ";
    cout << normdata->GetAvgLive(jhel)<<endl;
    cout << "Correction for non-synch of scalers = ";
    cout << normdata->GetCorrfact(jhel)<<endl;
  }

// This repeats what is in THaScaler::PrintSummary()
  printf("\n ------------ GmpNormAna Scaler Summary   ---------------- \n");
  Double_t clockrate = 1024;
  Double_t time_sec = myscaler->GetPulser("clock")/clockrate;
  printf("Time of run = %f sec \n",time_sec);

  return 1;
}

//_____________________________________________________________________________
  void GmpNormAna::WriteHist()
{
  for (UInt_t i = 0; i < hist.size(); i++) hist[i]->Write();
}

//_____________________________________________________________________________
  void GmpNormAna::BookHist()
{
// TDCs used for trigger latch pattern.  Used to tune cuts
// that define "simultaneity" of triggers.  
  
  Int_t nbin = 100;
  Float_t xlo = 0;
  Float_t xhi = 4000;
  char cname[50],ctitle[50];
  for (int i = 0; i < NTRIG; i++) {
    sprintf(cname,"Tlatch%d",i+1);
    sprintf(ctitle,"Trigger Latch TDC %d",i+1);
    hist.push_back(new TH1F(cname,ctitle,nbin,xlo,xhi));
  }

}


//_____________________________________________________________________________
THaAnalysisObject::EStatus GmpNormAna::Init( const TDatime& run_time ) 
{

  fStatus = kNotinit;
  MakePrefix();

  if (fDEBUG) *outputFile << "GmpNormAna:: Init !   the  name =  "<<fName<<"    "<<fSpectr<<endl<<flush;

  // Grab the scalers.
  // Of course these scalers are the event type 140 scaler data 
  THaAnalyzer* theAnalyzer = THaAnalyzer::GetInstance();
  TList* scalerList = theAnalyzer->GetScalers();
  TIter next(scalerList);
  while( THaScalerGroup* tscalgrp = static_cast<THaScalerGroup*>( next() )) {
    THaScaler *scaler = tscalgrp->GetScalerObj();
    if (CmpNoCase(fSpectr,scaler->GetName()) == 0) {
         myscaler = scaler;   // event type 140 data
         break;
    }
  }

  if (myscaler) {
    if (fDEBUG) *outputFile << "GmpNormAna:  Found the scalers"<<endl;
  } else {
    if (fDEBUG) *outputFile << "GmpNormAna:  Did NOT find the scalers"<<endl;
  }

  return fStatus = static_cast<EStatus>( SetupRawData( &run_time ) );


}



//_____________________________________________________________________________
Int_t GmpNormAna::Process(const THaEvData& evdata)
{
// Process the data:  Get prescale factors, get trigger
// latch pattern.  Process scalers.  All sorted by helicity
// if helicity was enabled.

  Int_t i, j, itrig;
  int fDEBUG=1;

  //  if (fDEBUG) cout << "GmpNormAna:  process the data ..."<<endl<<flush;

// helicity is -1,0,1.  here, 0 == irrespective of helicity.
// But if helicity disabled its always 0.
//  fHelEnable = evdata.HelicityEnabled();
//  Int_t helicity = evdata.GetHelicity();
  fHelEnable = kFALSE;
  Int_t helicity = 0;

  if (evdata.IsPhysicsTrigger()) {
      normdata->EvCount(helicity);
      itrig = evdata.GetEvType();
      normdata->TrigCount(itrig-1,helicity);  // Want to replace this with trigger latch -- later
  }

// If you have ~1000 events you probably have seen 
// prescale event; ok, this might not work for filesplitting
// if prescale event only goes into first file (sigh).
  
  if (!fSetPrescale && evdata.GetEvNum() > 1000) {
    fSetPrescale = kTRUE;
// CAREFUL, must GetPrescaleFactor(J) where J = 1,2,3 .. 
// (index starts at 1).  normdata starts at 0.
    if (fDEBUG) *outputFile << "Prescale factors.  Origin = "<<evdata.GetOrigPS()<<endl;
    for (i = 0; i < 8; i++) {
         if (fDEBUG) *outputFile << "     trig= "<<i+1<<"   ps= "<<evdata.GetPrescaleFactor(i+1)<<endl;
         normdata->SetPrescale(i,evdata.GetPrescaleFactor(i+1));
    }
    for (i = 8; i < 12; i++) normdata->SetPrescale(i,1);
  }

// TDC for trigger latch pattern.  The crate,slot,startchan 
// might change with experiments (I hope not).
//  It's either in roc5 (L-HRS) or roc6 (R-HRS)

   int crate,slot,startchan,iarm,index;

   int crateL = 5;
   int slotL = 16;
   int startchanL = 0;

   int crateR = 6;
   int slotR = 16;
   int startchanR = 64;
 
   if (fDEBUG) *outputFile << "fSpectr   "<<fSpectr<<endl;

   if (CmpNoCase(fSpectr,"Left") == 0) {
       iarm=0;
       crate = crateL;
       slot = slotL;
       startchan = startchanL;
   } else {
       iarm=1;
       crate = crateR;
       slot = slotR;
       startchan = startchanR;
   }

   if (fDEBUG) *outputFile << "arm, etc "<<iarm<<"   "<<crate<< "   "<<slot<< "   "<<startchan<<endl;

   for (i=0; i < NTRIG; i++) {
       for (j=0; j < MAXHIT; j++) {
           for (j = 0; j < 6; j++)  tdcdata[j*12+i] = 0;
       }
   }

   for (i = 0; i < 8; i++) {   // only 8 pre-scalable triggers 
        // map the channels to the triggers
        int tdctrig=0;
        if (iarm==0) {
  	    tdctrig = 8 - i;
        }
        if (iarm==1) {
  	   tdctrig = 72 - (startchan + i);
        }
         nhit[tdctrig-1] = evdata.GetNumHits(crate, slot, startchan+i);
         if (fDEBUG)  {
 	    if (nhit[tdctrig-1] > 0) {
	        *outputFile << "Hit on crate "<<crate<<"  Latch for TDC chan "<<startchan+i<<"   Nhit "<<nhit[tdctrig-1]<<"    Trigger "<<tdctrig<<endl;
	    }
	 }
         if (nhit[tdctrig-1] > 6) nhit[tdctrig-1] = 6;  

          for (j = 0; j < nhit[tdctrig-1]; j++) {
                 Int_t tdat = evdata.GetData(crate,slot,startchan+i,j);
                 index = j*NTRIG+tdctrig-1;
                 if (fDEBUG) *outputFile << "Indices "<<i<<"   "<<j<<"   "<<tdctrig<<"   "<<index<<"      Raw TDC "<<tdat<<endl;
		 if (index >= 0 && index < MAXHIT*NTRIG)  tdcdata[index] = tdat;
                 index = tdctrig-1;
                 if (index >= 0 && index < NTRIG) hist[index]->Fill(tdat);
	  }

    }

   TrigBits(helicity);

  if (myscaler) {
       if (myscaler->IsRenewed())  LiveTime();  
  }

  Int_t scaroc, header1,header2;
  if (iarm == 0)  {  // L_HRS
      scaroc=11;
      header1=0xabc30000; 
      header2=0xfed00017;     
  } else {
      scaroc=10;
      header1=0xceb10000; 
      header2=0xfed00017;     
  }
  if ( evdata.GetRocLength(scaroc)>20) {
     for (i = 0; i < evdata.GetRocLength(scaroc); i++) {
       Int_t sdat = evdata.GetRawData(scaroc,i);
       if(fDEBUG) *outputFile << "scaroc  on arm "<<fSpectr<<"   "<<i<<"   "<<sdat<<"   = 0x"<<hex<<sdat<<dec<<endl;
       if (sdat==header1) {
	 for (j=0; j<fgNumChan; j++) {
                roc_scaler[j]=evdata.GetRawData(scaroc,i+j);
                *outputFile << "Roc_scaler "<<j<<"   = 0x"<<hex<<roc_scaler[j]<<dec<<endl;
	 }
       }
       if (sdat==header2) {  /* TS DT scalers */
         TSLive1=evdata.GetRawData(scaroc,i+20);
         TSLive2=evdata.GetRawData(scaroc,i+21);
	 for (j=1; j<24; j++) {
	   if (fDEBUG) *outputFile << "TS DT scalers   "<<evdata.GetRawData(scaroc,i+j)<<endl;
	   }
       }
     }
  }

   TSdtime=100;
   if (TSLive1 != 0 && TSLive2 != 0) {
       Double_t x1 = 1.*TSLive1;
       Double_t x2 = 1.*TSLive2;
       TSdtime = 100.*(1.0- (x1/x2));
   }
   if (fDEBUG) *outputFile << "Trigger Supervisor  deadtime = "<<TSLive1<<"   "<<TSLive2<<"  "<<TSdtime<<endl;

   if (fDEBUG) Print();


  return 0;
}


//_____________________________________________________________________________
void GmpNormAna::Print( Option_t* ) const
{
  // Print for purpose of debugging.

   normdata->Print();

   PrintSummary();
}




void GmpNormAna::TrigBits( Int_t helicity ) {
// Figure out which triggers got a hit.  These are 
// multihit TDCs, so we need to sort out which hit 
// we want to take by applying cuts.

  int fDEBUG=1;

  int itrig,jhit,localtrig[NTRIG];

// These cuts can be decided by looking at the histograms.
// They importantly define "simultaneous" triggers, and 
// may therefore introduce a (small) systematic error

  static const Int_t cutlo = 500;
  static const Int_t cuthi = 1700;

  evtypebits = 0;
  memset(localtrig,0,NTRIG*sizeof(int));
  
  for (itrig = 0; itrig < NTRIG; itrig++) {   
    if (fDEBUG && nhit[itrig] > 0) {
           *outputFile << "TDC for trig "<<itrig+1;
           *outputFile << "   nhit = "<<nhit[itrig]<<endl;
    }
    for (jhit = 0; jhit < nhit[itrig]; jhit++) {
      Int_t tdat = tdcdata[NTRIG*jhit + itrig];
// Count at most once per event (don't double count)
// That's the purpose of localtrig
      if (fDEBUG) {
          *outputFile << "    TDC dat for hit ";
          *outputFile << jhit<<"   "<<"  index  "<<NTRIG*jhit + itrig<<"     data = "<<tdat<<endl;
      }
      if (tdat > cutlo && tdat < cuthi && localtrig[itrig]==0) {
          localtrig[itrig] = 1;
    	  normdata->TrigCount(itrig,helicity);
          evtypebits |= (2<<itrig);  
          if (fDEBUG) {
                *outputFile << "        trigger bit "<< itrig;
                *outputFile << "   helicity "<< helicity << "  count = ";
                *outputFile << normdata->GetTrigCount(itrig,helicity)<<endl;
	  }
      }
    }
  }

  if (fDEBUG) {
           *outputFile << "  trig bit pattern "<<evtypebits<<endl;
  }

}


//_____________________________________________________________________________

void GmpNormAna::LiveTime() {

// Compute the livetime by trigger type and by helicity.

  int fDEBUG=0;

   if (myscaler == 0) return;

   Int_t isrenewed=0;

 // Check if scalers have changed.
   for (Int_t i=0; i<8; i++) {
     ev140_scaler[i] = myscaler->GetTrig(0,i+1);
     if (ev140_scaler[i] != ev140_scaler[i+8]) isrenewed=1;
   }
 
   if ( !isrenewed) return;  // probably no new scaler data

   herecnt++;
   if (fDEBUG) *outputFile << "here count "<<herecnt<<endl;

   for (Int_t i=0; i<8; i++) {
     ev140_scaler[i+8] = ev140_scaler[i];
   }

   Int_t itrig, ihel, nhel, numtrig;
   Double_t trate, livetime;


  nhel = 1;  
  if (fHelEnable) nhel = 3;

  for (ihel = 0; ihel < nhel; ihel++) { 
    Int_t helicity = 0;
    if (ihel==1) helicity = -1;
    if (ihel==2) helicity = 1;

// CAREFUL ABOUT INDEX conventions
// THaScaler::GetTrig starts at itrig=1
// normdata indices all start at 0.

    for (itrig = 0; itrig < 12; itrig++) {

      numtrig = (Double_t)myscaler->GetTrig(helicity,itrig+1);
      trate = myscaler->GetTrigRate(helicity,itrig+1);
      if (fDEBUG) *outputFile << "trig= "<<itrig+1<<"   numtrig "<<numtrig<<"    rate =  "<<trate<<endl;

// Livetime by trigger type and by helicity:

      if (normdata->GetPrescale(itrig) > 0 && numtrig > 0) {
           livetime = normdata->GetTrigCount(itrig,helicity)*
                normdata->GetPrescale(itrig)/numtrig;
            normdata->SetLive(itrig, livetime, helicity); 

            if (fDEBUG) *outputFile << "Prescale "<<normdata->GetPrescale(itrig)<<"    "<<normdata->GetTrigCount(itrig,helicity)<<"    livetime =   "<<livetime<<endl;

            normdata->SetLive(itrig, livetime, helicity);


      }

    }

  }

   t1live = normdata->GetLive(0,0);
   if (fDEBUG) *outputFile  << "the  t1live = "<<t1live<<endl;

}


//_____________________________________________________________________________
ClassImp(GmpNormAna)

